@@ -12,7 +12,7 @@ allow_exact_list=
 allow_rate_limit=1
 allow_stream_buffer=1
 allow_compress=1
-compress_list="gzip|pigz|bzip2|pbzip2|xz|lzop|lz4"
+compress_list="gzip|pigz|bzip2|pbzip2|xz|lzop|lz4|zstd"
 
 # note that the backslash is NOT a metacharacter in a POSIX bracket expression!
 option_match='-[a-zA-Z0-9=-]+'   # matches short as well as long options
@@ -40,7 +40,7 @@ reject_and_die()
     local reason=$1
     log_cmd "auth.err" "btrbk REJECT" "$reason"
     echo "ERROR: ssh_filter_btrbk.sh: ssh command rejected: $reason: $SSH_ORIGINAL_COMMAND" 1>&2
-    exit 1
+    exit 255
 }
 
 run_cmd()
@@ -61,33 +61,32 @@ reject_filtered_cmd()
     fi
 
     if [[ -n "$allow_compress" ]]; then
-        decompress_match="((${compress_list}) -d -c( -[pT][0-9]+)? \| )?"
-        compress_match="( \| (${compress_list}) -c( -[0-9])?( -[pT][0-9]+)?)?"
+        decompress_match="(${compress_list}) -d -c( -[pT][0-9]+)?"
+        compress_match="(${compress_list}) -c( -[0-9])?( -[pT][0-9]+)?"
     else
         decompress_match=
         compress_match=
     fi
 
-    if [[ -n "$allow_stream_buffer" ]]; then
-        stream_buffer_match="(mbuffer -q -m [0-9]+[kmg]? \| )?"
+    # rate_limit_remote and stream_buffer_remote use combined
+    # "mbuffer" as of btrbk-0.29.0
+    if [[ -n "$allow_stream_buffer" ]] || [[ -n "$allow_rate_limit" ]]; then
+        mbuffer_match="mbuffer -v 1 -q( -s [0-9]+[kmgKMG]?)?( -m [0-9]+[kmgKMG]?)?( -[rR] [0-9]+[kmgtKMGT]?)?"
     else
-        stream_buffer_match=
-    fi
-
-    if [[ -n "$allow_rate_limit" ]]; then
-        rate_limit_match="( \| pv -q -L [0-9]+[kmgt]?)?"
-    else
-        rate_limit_match=
+        mbuffer_match=
     fi
 
     # allow multiple paths (e.g. "btrfs subvolume snapshot <src> <dst>")
-    btrfs_cmd_match="^${decompress_match}${stream_buffer_match}(${allow_list})( ${option_match})*( ${path_match})+${compress_match}${rate_limit_match}$"
+    allow_cmd_match="(${allow_list})( ${option_match})*( ${path_match})+"
+    stream_in_match="(${decompress_match} \| )?(${mbuffer_match} \| )?"
+    stream_out_match="( \| ${mbuffer_match})?( \| ${compress_match}$)?"
 
-    if [[ $SSH_ORIGINAL_COMMAND =~ $btrfs_cmd_match ]] ; then
+    allow_stream_match="^${stream_in_match}${allow_cmd_match}${stream_out_match}"
+    if [[ $SSH_ORIGINAL_COMMAND =~ $allow_stream_match ]] ; then
         return 0
     fi
 
-    exact_cmd_match="^${allow_exact_list}$";
+    exact_cmd_match="^(${allow_exact_list})$";
     if [[ $SSH_ORIGINAL_COMMAND =~ $exact_cmd_match ]] ; then
         return 0
     fi
@@ -126,6 +125,7 @@ while [[ "$#" -ge 1 ]]; do
 
       -t|--target)
           allow_cmd "${sudo_prefix}btrfs receive"
+          allow_cmd "${sudo_prefix}mkdir"
           ;;
 
       -c|--compress)
@@ -155,17 +155,19 @@ while [[ "$#" -ge 1 ]]; do
 
       *)
           echo "ERROR: ssh_filter_btrbk.sh: failed to parse command line option: $key" 1>&2
-          exit 1
+          exit 255
           ;;
     esac
     shift
 done
 
-allow_cmd "${sudo_prefix}btrfs subvolume show"; # subvolume queries are always allowed
-allow_exact_cmd "${sudo_prefix}btrfs subvolume list ${file_match}"; # subvolume queries are always allowed
-allow_cmd "${sudo_prefix}readlink"              # used to resolve mountpoints
-allow_exact_cmd "cat /proc/self/mountinfo"      # used to resolve mountpoints
-allow_exact_cmd "cat /proc/self/mounts"         # legacy, for btrbk < 0.27.0
+# NOTE: subvolume queries are NOT affected by "--restrict-path":
+# btrbk also calls show/list on the mount point of the subvolume
+allow_exact_cmd "${sudo_prefix}btrfs subvolume (show|list)( ${option_match})* ${file_match}";
+allow_cmd "${sudo_prefix}readlink"                    # resolve symlink
+allow_exact_cmd "${sudo_prefix}test -d ${file_match}" # check directory (only for compat=busybox)
+allow_exact_cmd "cat /proc/self/mountinfo"            # resolve mountpoints
+allow_exact_cmd "cat /proc/self/mounts"               # legacy, for btrbk < 0.27.0
 
 # remove leading "|" on alternation lists
 allow_list=${allow_list#\|}
@@ -173,16 +175,16 @@ allow_exact_list=${allow_exact_list#\|}
 restrict_path_list=${restrict_path_list#\|}
 
 case "$SSH_ORIGINAL_COMMAND" in
-    *\.\./*)  reject_and_die "directory traversal"  ;;
-    *\$*)     reject_and_die "unsafe character"     ;;
-    *\&*)     reject_and_die "unsafe character"     ;;
-    *\(*)     reject_and_die "unsafe character"     ;;
-    *\{*)     reject_and_die "unsafe character"     ;;
-    *\;*)     reject_and_die "unsafe character"     ;;
-    *\<*)     reject_and_die "unsafe character"     ;;
-    *\>*)     reject_and_die "unsafe character"     ;;
-    *\`*)     reject_and_die "unsafe character"     ;;
-    *\|*)     [[ -n "$allow_compress" ]] || [[ -n "$allow_rate_limit" ]] || [[ -n "$allow_stream_buffer" ]] || reject_and_die "unsafe character (compression disallowed)" ;;
+    *\.\./*)  reject_and_die 'directory traversal'  ;;
+    *\$*)     reject_and_die 'unsafe character "$"' ;;
+    *\&*)     reject_and_die 'unsafe character "&"' ;;
+    *\(*)     reject_and_die 'unsafe character "("' ;;
+    *\{*)     reject_and_die 'unsafe character "{"' ;;
+    *\;*)     reject_and_die 'unsafe character ";"' ;;
+    *\<*)     reject_and_die 'unsafe character "<"' ;;
+    *\>*)     reject_and_die 'unsafe character ">"' ;;
+    *\`*)     reject_and_die 'unsafe character "`"' ;;
+    *\|*)     [[ -n "$allow_compress" ]] || [[ -n "$allow_rate_limit" ]] || [[ -n "$allow_stream_buffer" ]] || reject_and_die 'unsafe character "|"' ;;
 esac
 
 reject_filtered_cmd
